# License: MIT
# Copyright © 2024 Frequenz Energy-as-a-Service GmbH

"""The Historical Forecast Iterator."""

from __future__ import annotations

from collections.abc import AsyncIterator
from datetime import datetime

from frequenz.api.common.v1.pagination import pagination_params_pb2
from frequenz.api.weather import weather_pb2, weather_pb2_grpc
from google.protobuf import timestamp_pb2

from ._types import ForecastFeature, HistoricalForecasts, Location

DEFAULT_PAGE_SIZE = 20
EMPTY_PAGE_TOKEN = ""


# pylint: disable=too-many-instance-attributes
class HistoricalForecastIterator(AsyncIterator[HistoricalForecasts]):
    """An iterator over historical weather forecasts."""

    def __init__(  # pylint: disable=too-many-arguments, too-many-positional-arguments
        self,
        stub: weather_pb2_grpc.WeatherForecastServiceAsyncStub,
        locations: list[Location],
        features: list[ForecastFeature],
        start: datetime,
        end: datetime,
        page_size: int = DEFAULT_PAGE_SIZE,
    ) -> None:
        """Initialize the iterator.

        Args:
            stub: The gRPC stub to use for communication with the API.
            locations: Locations to get historical weather forecasts for.
            features: Features to get historical weather forecasts for.
            start: Start of the time range to get historical weather forecasts for.
            end: End of the time range to get historical weather forecasts for.
            page_size: The number of historical weather forecasts to get per page.
        """
        self._stub = stub
        self.locations = locations
        self.features = features

        self.start_ts = timestamp_pb2.Timestamp()
        self.start_ts.FromDatetime(start)
        self.end_ts = timestamp_pb2.Timestamp()
        self.end_ts.FromDatetime(end)

        self.location_forecasts: list[weather_pb2.LocationForecast] = []
        self.page_token: str | None = None
        self.page_size = page_size

    def __aiter__(self) -> "HistoricalForecastIterator":
        """Return the iterator.

        Returns:
            The iterator.
        """
        return self

    async def __anext__(self) -> HistoricalForecasts:
        """Get the next historical weather forecast.

        Returns:
            The next historical weather forecast.

        Raises:
            StopAsyncIteration: If there are no more historical weather forecasts.
        """
        if self.page_token == EMPTY_PAGE_TOKEN:
            raise StopAsyncIteration

        pagination_params = pagination_params_pb2.PaginationParams()
        pagination_params.page_size = self.page_size
        if self.page_token is not None:
            pagination_params.page_token = self.page_token

        response: weather_pb2.GetHistoricalWeatherForecastResponse = (
            await self._stub.GetHistoricalWeatherForecast(
                weather_pb2.GetHistoricalWeatherForecastRequest(
                    locations=(location.to_pb() for location in self.locations),
                    features=(feature.value for feature in self.features),
                    start_ts=self.start_ts,
                    end_ts=self.end_ts,
                    pagination_params=pagination_params,
                )
            )
        )

        if (
            response.pagination_info is None
            or response.pagination_info.next_page_token is None
        ):
            raise StopAsyncIteration

        self.page_token = response.pagination_info.next_page_token
        if len(response.location_forecasts) == 0:
            raise StopAsyncIteration

        return HistoricalForecasts.from_pb(response)



# License: MIT
# Copyright © 2023 Frequenz Energy-as-a-Service GmbH

"""Types used by the Weather Forecast API client."""

from __future__ import annotations  # required for constructor type hinting

import datetime as dt
import enum
import logging
import typing
from collections import namedtuple
from dataclasses import dataclass

import numpy as np
from frequenz.api.common.v1 import location_pb2
from frequenz.api.weather import weather_pb2

# Set up logging
_logger = logging.getLogger(__name__)


class ForecastFeature(enum.Enum):
    """Weather forecast features available through the API."""

    UNSPECIFIED = weather_pb2.ForecastFeature.FORECAST_FEATURE_UNSPECIFIED
    """Unspecified forecast feature."""

    TEMPERATURE_2_METRE = (
        weather_pb2.ForecastFeature.FORECAST_FEATURE_TEMPERATURE_2_METRE
    )
    """Temperature at 2m above the earth's surface."""

    U_WIND_COMPONENT_100_METRE = (
        weather_pb2.ForecastFeature.FORECAST_FEATURE_U_WIND_COMPONENT_100_METRE
    )
    """Eastward wind component at 100m altitude."""

    V_WIND_COMPONENT_100_METRE = (
        weather_pb2.ForecastFeature.FORECAST_FEATURE_V_WIND_COMPONENT_100_METRE
    )
    """Northward wind component at 100m altitude."""

    U_WIND_COMPONENT_10_METRE = (
        weather_pb2.ForecastFeature.FORECAST_FEATURE_U_WIND_COMPONENT_10_METRE
    )
    """Eastward wind component at 10m altitude."""

    V_WIND_COMPONENT_10_METRE = (
        weather_pb2.ForecastFeature.FORECAST_FEATURE_V_WIND_COMPONENT_10_METRE
    )
    """Northward wind component at 10m altitude."""

    SURFACE_SOLAR_RADIATION_DOWNWARDS = (
        weather_pb2.ForecastFeature.FORECAST_FEATURE_SURFACE_SOLAR_RADIATION_DOWNWARDS
    )
    """Surface solar radiation downwards."""

    SURFACE_NET_SOLAR_RADIATION = (
        weather_pb2.ForecastFeature.FORECAST_FEATURE_SURFACE_NET_SOLAR_RADIATION
    )
    """Surface net solar radiation."""

    @classmethod
    def from_pb(
        cls, forecast_feature: weather_pb2.ForecastFeature.ValueType
    ) -> ForecastFeature:
        """Convert a protobuf ForecastFeature value to ForecastFeature enum.

        Args:
            forecast_feature: protobuf forecast feature to convert.

        Returns:
            Enum value corresponding to the protobuf message.
        """
        if not any(t.value == forecast_feature for t in ForecastFeature):
            _logger.warning(
                "Unknown forecast feature %s. Returning UNSPECIFIED.", forecast_feature
            )
            return cls.UNSPECIFIED

        return ForecastFeature(forecast_feature)


SOLAR_PARAMETERS = {
    ForecastFeature.SURFACE_SOLAR_RADIATION_DOWNWARDS,
    ForecastFeature.SURFACE_NET_SOLAR_RADIATION,
}


@dataclass(frozen=True)
class Location:
    """Location data.

    Attributes:
        latitude: latitude of the location.
        longitude: longitude of the location.
        country_code: ISO 3166-1 alpha-2 country code of the location.
    """

    latitude: float
    longitude: float
    country_code: str

    @classmethod
    def from_pb(cls, location: location_pb2.Location) -> Location:
        """Convert a protobuf Location message to Location object.

        Args:
            location: protobuf location to convert.

        Returns:
            Location object corresponding to the protobuf message.
        """
        return cls(
            latitude=location.latitude,
            longitude=location.longitude,
            country_code=location.country_code,
        )

    def to_pb(self) -> location_pb2.Location:
        """Convert a Location object to protobuf Location message.

        Returns:
            Protobuf message corresponding to the Location object.
        """
        return location_pb2.Location(
            latitude=self.latitude,
            longitude=self.longitude,
            country_code=self.country_code,
        )


@dataclass(frozen=True)
class Forecasts:
    """Weather forecast data."""

    _forecasts_pb: weather_pb2.ReceiveLiveWeatherForecastResponse

    @classmethod
    def from_pb(
        cls, forecasts: weather_pb2.ReceiveLiveWeatherForecastResponse
    ) -> Forecasts:
        """Convert a protobuf Forecast message to Forecast object.

        Args:
            forecasts: protobuf message with live forecast data.

        Returns:
            Forecast object corresponding to the protobuf message.
        """
        return cls(_forecasts_pb=forecasts)

    # pylint: disable=too-many-locals,too-many-branches,too-many-statements
    def to_ndarray_vlf(
        self,
        validity_times: list[dt.datetime] | None = None,
        locations: list[Location] | None = None,
        features: list[ForecastFeature] | None = None,
    ) -> np.ndarray[
        # the shape is known to be 3 dimensional, but the length of each dimension is
        # not fixed, so we use typing.Any, instead of the usual const generic
        # parameters.
        tuple[typing.Any, typing.Any, typing.Any],
        np.dtype[np.float64],
    ]:
        """Convert a Forecast object to numpy array and use NaN to mark irrelevant data.

        If any of the filters are None, all values for that parameter will be returned.

        Args:
            validity_times: The validity times to filter by.
            locations: The locations to filter by.
            features: The features to filter by.

        Returns:
            Numpy array of shape (num_validity_times, num_locations, num_features)

        Raises:
            ValueError: If the forecasts data is missing or invalid.
            RuntimeError: If there is an error processing the forecast data.
        """
        # check for empty forecasts data
        if not self._forecasts_pb.location_forecasts:
            raise ValueError("Forecast data is missing or invalid.")

        try:
            num_times = len(self._forecasts_pb.location_forecasts[0].forecasts)
            num_locations = len(self._forecasts_pb.location_forecasts)
            num_features = len(
                self._forecasts_pb.location_forecasts[0].forecasts[0].features
            )

            # Look for the proto indexes of the filtered times, locations and features
            location_indexes = []
            validity_times_indexes = []
            feature_indexes = []

            # get the location indexes of the proto for the filtered locations
            if locations:
                for location in locations:
                    found = False
                    for l_index, location_forecast in enumerate(
                        self._forecasts_pb.location_forecasts
                    ):
                        if location == Location.from_pb(location_forecast.location):
                            location_indexes.append(l_index)
                            found = True
                            break
                    if not found:
                        # remember position of missing location
                        location_indexes.append(-1)
            else:
                location_indexes = list(range(num_locations))

            # get the val indexes of the proto for the filtered validity times
            if validity_times:
                for req_validitiy_time in validity_times:
                    found = False
                    for t_index, val_time in enumerate(
                        self._forecasts_pb.location_forecasts[0].forecasts
                    ):
                        if req_validitiy_time == val_time.valid_at_ts.ToDatetime():
                            validity_times_indexes.append(t_index)
                            found = True
                            break
                    if not found:
                        # remember position of missing validity time
                        validity_times_indexes.append(-1)
            else:
                validity_times_indexes = list(range(num_times))

            # get the feature indexes of the proto for the filtered features
            if features:
                for req_feature in features:
                    found = False
                    for f_index, feature in enumerate(
                        self._forecasts_pb.location_forecasts[0].forecasts[0].features
                    ):
                        if req_feature == ForecastFeature.from_pb(feature.feature):
                            feature_indexes.append(f_index)
                            found = True
                            break
                    if not found:
                        # remember position of missing feature
                        feature_indexes.append(-1)
            else:
                feature_indexes = list(range(num_features))

            array = np.full(
                (
                    len(validity_times_indexes),
                    len(location_indexes),
                    len(feature_indexes),
                ),
                np.nan,
            )

            array_l_index = 0

            for l_index in location_indexes:
                array_t_index = 0

                for t_index in validity_times_indexes:
                    array_f_index = 0

                    for f_index in feature_indexes:
                        # This fails if there was data missing for at least one of the
                        # keys and we don't update the array but leave it as NaN
                        if l_index >= 0 and t_index >= 0 and f_index >= 0:
                            array[array_t_index, array_l_index, array_f_index] = (
                                self._forecasts_pb.location_forecasts[l_index]
                                .forecasts[t_index]
                                .features[f_index]
                                .value
                            )
                        array_f_index += 1

                    array_t_index += 1

                array_l_index += 1

            # Check if the array shape matches the number of filtered times, locations
            # and features
            if validity_times is not None and array.shape[0] != len(validity_times):
                raise ValueError(
                    f"The count of validity times in the array({array.shape[0]}) does "
                    f"not match the requested validity times count ({validity_times}."
                )
            if locations is not None and array.shape[1] != len(locations):
                raise ValueError(
                    f"The count of location in the array ({array.shape[1]}) does not "
                    f"match the requested location count ({locations})."
                )
            if features is not None and array.shape[2] != len(features):
                raise ValueError(
                    f"The count of features in the array ({array.shape[2]}) does not "
                    f"match the requested feature count ({features})."
                )

        # catch all exceptions
        except Exception as e:
            raise RuntimeError("Error processing forecast data") from e

        return array

    # pylint: disable= too-many-arguments, too-many-positional-arguments
    def upsample_vlf(
        self,
        array: np.ndarray[
            tuple[typing.Any, typing.Any, typing.Any], np.dtype[np.float64]
        ],
        validity_times: list[dt.datetime],
        target_times: list[dt.datetime],
        features: list[ForecastFeature],
        solar_offset_sec: int = 1800,
    ) -> np.ndarray[tuple[typing.Any, typing.Any, typing.Any], np.dtype[np.float64]]:
        """Upsample the forecast array to requested timestamps.

        Args:
            array: 3D array from to_ndarray_vlf (time, location, feature)
            validity_times: List of original timestamps
            target_times: List of desired timestamps to interpolate to
            features: List of forecast features
            solar_offset_sec: Time offset in seconds to shift solar forecasts

        Returns:
            Resampled 3D array with same structure interpolated to target timestamps

        Raises:
            ValueError: If input dimensions don't match or timestamps aren't monotonic
        """
        # Check input dimensions
        if array.shape[0] != len(validity_times):
            raise ValueError(
                f"Time dimension of input array ({array.shape[0]}) does not match "
                f"number of validity times ({len(validity_times)})"
            )
        if array.shape[2] != len(features):
            raise ValueError(
                f"Feature dimension of input array ({array.shape[2]}) does not match "
                f"number of features ({len(features)})"
            )

        # Validate target timestamps are strictly increasing
        if not all(t1 < t2 for t1, t2 in zip(target_times[:-1], target_times[1:])):
            raise ValueError("target_times must be strictly increasing")

        vts = np.array([t.timestamp() for t in validity_times])
        tts = np.array([t.timestamp() for t in target_times])

        resampled = np.zeros((len(target_times), array.shape[1], array.shape[2]))

        # Get indices of solar and non-solar features
        solar_idxs = [i for i, f in enumerate(features) if f in SOLAR_PARAMETERS]
        other_idxs = [i for i, f in enumerate(features) if f not in SOLAR_PARAMETERS]

        # Handle non-solar features with direct interpolation
        if other_idxs:
            resampled[..., other_idxs] = np.apply_along_axis(
                lambda x: np.interp(tts, vts, x),
                axis=0,
                arr=array[..., other_idxs],
            )

        # Handle solar features with shifted interpolation
        if solar_idxs:
            # Shift validity times by solar_offset_sec for solar parameters
            shifted_vts = vts + solar_offset_sec
            resampled[..., solar_idxs] = np.apply_along_axis(
                lambda x: np.interp(
                    tts,
                    shifted_vts,
                    x,
                    left=x[0],
                    right=x[-1],
                ),
                axis=0,
                arr=array[..., solar_idxs],
            )

        return resampled

    def flatten(self) -> list[ForecastData]:
        """Flatten forecast data into a list of ForecastData tuples.

        Returns:
            List of ForecastData tuples containing the flattened forecast data.

        Raises:
            ValueError: If the forecasts data is missing or invalid.
        """
        # check for empty forecasts data
        if not self._forecasts_pb.location_forecasts:
            raise ValueError("Forecast data is missing or invalid.")

        return flatten(list(self._forecasts_pb.location_forecasts))


ForecastData = namedtuple(
    "ForecastData",
    ["creation_ts", "latitude", "longitude", "validity_ts", "feature", "value"],
)


@dataclass(frozen=True)
class HistoricalForecasts:
    """Historical weather forecast data."""

    _forecasts_pb: weather_pb2.GetHistoricalWeatherForecastResponse

    @classmethod
    def from_pb(
        cls, forecasts: weather_pb2.GetHistoricalWeatherForecastResponse
    ) -> HistoricalForecasts:
        """Convert a protobuf Forecast message to Forecast object.

        Args:
            forecasts: protobuf message with historical forecast data.

        Returns:
            Forecast object corresponding to the protobuf message.
        """
        return cls(_forecasts_pb=forecasts)

    def flatten(
        self,
    ) -> list[ForecastData]:
        """Flatten a Forecast object to a list of named tuples of data.

        Returns:
            List of named tuples with the flattened forecast data.

        Raises:
            ValueError: If the forecasts data is missing or invalid.
        """
        # check for empty forecasts data
        if not self._forecasts_pb.location_forecasts:
            raise ValueError("Forecast data is missing or invalid.")

        return flatten(list(self._forecasts_pb.location_forecasts))


def flatten(
    location_forecasts: list[weather_pb2.LocationForecast],
) -> list[ForecastData]:
    """Flatten a Forecast object to a list of named tuples of data.

    Each tuple contains the following data:
    - creation timestamp
    - latitude
    - longitude
    - validity timestamp
    - feature
    - forecast value

    Args:
        location_forecasts: The location forecasts to flatten.

    Returns:
        List of named tuples with the flattened forecast data.
    """
    data = []
    for location_forecast in location_forecasts:
        for forecasts in location_forecast.forecasts:
            for feature_forecast in forecasts.features:
                # Create and append an instance of the named tuple instead of a plain tuple
                data.append(
                    ForecastData(
                        creation_ts=location_forecast.creation_ts.ToDatetime(),
                        latitude=location_forecast.location.latitude,
                        longitude=location_forecast.location.longitude,
                        validity_ts=forecasts.valid_at_ts.ToDatetime(),
                        feature=ForecastFeature(feature_forecast.feature),
                        value=feature_forecast.value,
                    )
                )

    return data



# License: MIT
# Copyright © 2023 Frequenz Energy-as-a-Service GmbH

"""The Weather Forecast API client."""

from __future__ import annotations

from datetime import datetime

from frequenz.api.weather import weather_pb2, weather_pb2_grpc
from frequenz.channels import Receiver
from frequenz.client.base.channel import ChannelOptions
from frequenz.client.base.client import BaseApiClient
from frequenz.client.base.exception import ClientNotConnected
from frequenz.client.base.streaming import GrpcStreamBroadcaster

from ._historical_forecast_iterator import HistoricalForecastIterator
from ._types import ForecastFeature, Forecasts, Location


class Client(BaseApiClient[weather_pb2_grpc.WeatherForecastServiceStub]):
    """Weather forecast client."""

    def __init__(
        self,
        server_url: str,
        *,
        connect: bool = True,
        channel_defaults: ChannelOptions = ChannelOptions(),
    ) -> None:
        """Initialize the client.

        Args:
            server_url: The URL of the server to connect to.
            connect: Whether to connect to the server as soon as a client instance is
                created. If `False`, the client will not connect to the server until
                [connect()][frequenz.client.base.client.BaseApiClient.connect] is
                called.
            channel_defaults: Default options for the gRPC channel.
        """
        super().__init__(
            server_url,
            weather_pb2_grpc.WeatherForecastServiceStub,
            connect=connect,
            channel_defaults=channel_defaults,
        )
        self._streams: dict[
            tuple[Location | ForecastFeature, ...],
            GrpcStreamBroadcaster[
                weather_pb2.ReceiveLiveWeatherForecastResponse, Forecasts
            ],
        ] = {}

    @property
    def stub(self) -> weather_pb2_grpc.WeatherForecastServiceAsyncStub:
        """The gRPC stub for the API.

        Returns:
            The async gRPC stub for the Weather Forecast Service.

        Raises:
            ClientNotConnected: If the client is not connected to the server.
        """
        if self.channel is None or self._stub is None:
            raise ClientNotConnected(server_url=self.server_url, operation="stub")
        # This type: ignore is needed because we need to cast the sync stub to
        # the async stub, but we can't use cast because the async stub doesn't
        # actually exists to the eyes of the interpreter, it only exists for the
        # type-checker, so it can only be used for type hints.
        return self._stub  # type: ignore

    async def stream_live_forecast(
        self,
        locations: list[Location],
        features: list[ForecastFeature],
    ) -> Receiver[Forecasts]:
        """Stream live weather forecast data.

        Args:
            locations: locations to stream data for.
            features: features to stream data for.

        Returns:
            A channel receiver for weather forecast data.
        """
        stream_key = tuple(tuple(locations) + tuple(features))

        if stream_key not in self._streams:
            self._streams[stream_key] = GrpcStreamBroadcaster(
                f"weather-forecast-{stream_key}",
                lambda: self.stub.ReceiveLiveWeatherForecast(
                    weather_pb2.ReceiveLiveWeatherForecastRequest(
                        locations=(location.to_pb() for location in locations),
                        features=(feature.value for feature in features),
                    )
                ),
                Forecasts.from_pb,
            )
        return self._streams[stream_key].new_receiver()

    def hist_forecast_iterator(
        self,
        locations: list[Location],
        features: list[ForecastFeature],
        start: datetime,
        end: datetime,
    ) -> HistoricalForecastIterator:
        """Stream historical weather forecast data.

        Args:
            locations: locations to stream data for.
            features: features to stream data for.
            start: start of the time range to stream data for.
            end: end of the time range to stream data for.

        Returns:
            A channel receiver for weather forecast data.
        """
        return HistoricalForecastIterator(self.stub, locations, features, start, end)



# License: MIT
# Copyright © 2025 Frequenz Energy-as-a-Service GmbH

"""CLI tool to iterate over historical weather forecast data and print it in CSV format."""

import argparse
import asyncio
from datetime import datetime, timedelta

from frequenz.client.base.channel import ChannelOptions, KeepAliveOptions, SslOptions

from frequenz.client.weather._client import Client
from frequenz.client.weather._types import ForecastFeature, Location


def main() -> None:
    """Parse arguments and run the client."""
    parser = argparse.ArgumentParser()
    parser.add_argument(
        "--url",
        type=str,
        help="URL of the Weather service",
    )
    parser.add_argument(
        "--mode",
        type=str,
        choices=["historical", "live"],
        required=True,
        help="Operation mode: historical or live forecasts",
    )
    parser.add_argument(
        "--feature",
        type=str,
        nargs="+",
        choices=[e.name for e in ForecastFeature],
        help="Feature names",
        required=True,
    )
    parser.add_argument(
        "--location",
        type=lambda s: tuple(map(float, s.split(","))),  # One-liner lambda
        required=True,
        help='Location in lat,lon format (e.g., "37.7749,-122.4194")',
    )

    parser.add_argument(
        "--start",
        type=datetime.fromisoformat,
        help="Start datetime in YYYY-MM-DDTHH:MM:SS format (historical mode only)",
    )
    parser.add_argument(
        "--end",
        type=datetime.fromisoformat,
        help="End datetime in YYYY-MM-DDTHH:MM:SS format (historical mode only)",
    )

    args = parser.parse_args()

    # Validate arguments based on mode
    if args.mode == "historical":
        if args.start is None or args.end is None:
            parser.error("--start and --end are required in historical mode")
    else:
        if args.start is not None or args.end is not None:
            parser.error("--start and --end can only be used in historical mode")

    try:
        asyncio.run(
            run(
                service_address=args.url,
                location=args.location,
                feature_names=args.feature,
                start=args.start,
                end=args.end,
                mode=args.mode,
            )
        )
    except KeyboardInterrupt:
        print("\nReceived interrupt, shutting down...")


async def run(  # pylint: disable=too-many-arguments
    *,
    service_address: str,
    location: tuple[float, float],
    feature_names: list[str],
    start: datetime,
    end: datetime,
    mode: str,
) -> None:
    """Run the client.

    Args:
        service_address: service address
        location: location in lat, lon format
        feature_names: feature names
        start: start datetime (historical mode only)
        end: end datetime (historical mode only)
        mode: operation mode ("historical" or "live")
    """
    client = Client(
        service_address,
        channel_defaults=ChannelOptions(
            ssl=SslOptions(enabled=False),
            keep_alive=KeepAliveOptions(
                enabled=True,
                timeout=timedelta(minutes=5),
                interval=timedelta(seconds=20),
            ),
        ),
    )

    features = [ForecastFeature[feature_name] for feature_name in feature_names]
    locations = [
        Location(
            latitude=location[0],
            longitude=location[1],
            country_code="",
        ),
    ]

    print("creation_ts,validity_ts,latitude,longitude,feature,value")

    if mode == "historical":
        location_forecast_iterator = client.hist_forecast_iterator(
            features=features, locations=locations, start=start, end=end
        )
        async for forecasts in location_forecast_iterator:
            for fc in forecasts.flatten():
                row = (
                    fc.creation_ts,
                    fc.validity_ts,
                    fc.latitude,
                    fc.longitude,
                    fc.feature,
                    fc.value,
                )
                print(",".join(str(e) for e in row))
    else:
        receiver = await client.stream_live_forecast(
            locations=locations,
            features=features,
        )
        while True:
            live_forecasts = await receiver.receive()
            for fc in live_forecasts.flatten():
                row = (
                    fc.creation_ts,
                    fc.validity_ts,
                    fc.latitude,
                    fc.longitude,
                    fc.feature,
                    fc.value,
                )
                print(",".join(str(e) for e in row))
            print("--------Waiting for next forecast---------")


if __name__ == "__main__":
    main()
